@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2014 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -224,82 +224,182 @@ public final class SmalMessageStorage extends AbstractSMALStorage implements IMa
     }
 
     @Override
-    public MailMessage[] searchMessages(final String folder, final IndexRange indexRange, final MailSortField sortField, final OrderDirection order, final SearchTerm<?> searchTerm, final MailField[] fields) throws OXException {
+    public MailMessage[] searchMessages(String folder, IndexRange indexRange, MailSortField sortField, OrderDirection order, SearchTerm<?> searchTerm, MailField[] fields) throws OXException {
+        return searchMessages(folder, indexRange, sortField, order, searchTerm, fields, null);
+    }
+
+    @Override
+    public MailMessage[] searchMessages(String folder, IndexRange indexRange, MailSortField sortField, OrderDirection order, SearchTerm<?> searchTerm, MailField[] fields, String[] headerNames) throws OXException {
         final IndexFacadeService indexFacade = getIndexFacadeService();
         if (searchTerm == null || indexFacade == null || isBlacklisted() || !isIndexingAllowed()) {
-            return smalMailAccess.getDelegateMailAccess().getMessageStorage().searchMessages(folder, indexRange, sortField, order, searchTerm, fields);
+            IMailMessageStorage messageStorage = smalMailAccess.getDelegateMailAccess().getMessageStorage();
+            if (messageStorage instanceof IMailMessageStorageExt) {
+                IMailMessageStorageExt storageExt = (IMailMessageStorageExt) messageStorage;
+                return storageExt.searchMessages(folder, indexRange, sortField, order, searchTerm, fields, headerNames);
+            }
+
+            // Load headers manually
+            if (null == headerNames || headerNames.length <= 0) {
+                return messageStorage.searchMessages(folder, indexRange, sortField, order, searchTerm, fields);
+            }
+
+            MailField[] fieldsToUse = fields;
+            fieldsToUse = MailFields.addIfAbsent(fieldsToUse, MailField.ID);
+            MailMessage[] results = messageStorage.searchMessages(folder, indexRange, sortField, order, searchTerm, fieldsToUse);
+            if (null == results || results.length <= 0) {
+                return results;
+            }
+
+            int length = results.length;
+            MailMessage[] headers;
+            {
+                String[] ids = new String[length];
+                for (int i = ids.length; i-- > 0;) {
+                    MailMessage m = results[i];
+                    ids[i] = null == m ? null : m.getMailId();
+                }
+                headers = messageStorage.getMessages(folder, ids, MailFields.toArray(MailField.HEADERS));
+            }
+
+            for (int i = length; i-- > 0;) {
+                MailMessage mailMessage = results[i];
+                if (null != mailMessage) {
+                    MailMessage header = headers[i];
+                    if (null != header) {
+                        for (String headerName : headerNames) {
+                            String[] values = header.getHeader(headerName);
+                            if (null != values) {
+                                for (String value : values) {
+                                    mailMessage.addHeader(headerName, value);
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+
+            return results;
         }
 
-        // Close for the time accessing the index
-        smalMailAccess.closeDelegateMailAccess();
-        // Access index
-        IndexAccess<MailMessage> indexAccess = null;
-        try {
-            final MailFields mfs = new MailFields(fields);
-            indexAccess = indexFacade.acquireIndexAccess(Types.EMAIL, session);
-            final boolean isIndexed = indexAccess.isIndexed(String.valueOf(accountId), folder);
-            if (isIndexed && MailUtility.getIndexableFields(indexAccess).containsAll(mfs)) {
-                final AccountFolders accountFolders = new AccountFolders(String.valueOf(accountId), Collections.singleton(folder));
-                final QueryParameters.Builder builder = new QueryParameters.Builder().setAccountFolders(Collections.singleton(accountFolders));
-
-                if (sortField != null) {
-                    final MailField field = MailField.getField(sortField.getField());
-                    final MailIndexField indexSortField = MailIndexField.getFor(field);
-                    if (indexSortField != null) {
-                        builder.setSortField(indexSortField);
-                    }
-
-                    if (order != null) {
-                        builder.setOrder(order == OrderDirection.ASC ? Order.ASC : Order.DESC);
-                    }
-                }
-
-                final QueryParameters parameters = builder.setHandler(SearchHandlers.CUSTOM).setSearchTerm(searchTerm).build();
-                final IndexResult<MailMessage> result = indexAccess.query(parameters, MailIndexField.getFor(fields));
-
-                List<IndexDocument<MailMessage>> documents = result.getResults();
-                List<MailMessage> mails;
-                if (indexRange != null) {
-                    final int fromIndex = indexRange.start;
-                    int toIndex = indexRange.end;
-                    if ((documents == null) || documents.isEmpty()) {
-                        mails = Collections.emptyList();
-                    }
-                    if ((fromIndex) > documents.size()) {
-                        /*
-                         * Return empty iterator if start is out of range
-                         */
-                        mails = Collections.emptyList();
-                    }
-                    /*
-                     * Reset end index if out of range
-                     */
-                    if (toIndex >= documents.size()) {
-                        toIndex = documents.size();
-                    }
-                    documents = documents.subList(fromIndex, toIndex);
-                }
-
-                mails = IndexDocumentHelper.messagesFrom(documents);
-                return mails.toArray(new MailMessage[mails.size()]);
-            }
-        } catch (final Throwable t) {
-            ExceptionUtils.handleThrowable(t);
-            LOG.warn("Index search failed. Falling back to message storage.", t);
-        } finally {
-            if (indexAccess != null) {
-                indexFacade.releaseIndexAccess(indexAccess);
-            }
+        if (null == headerNames || headerNames.length <= 0) { // We cannot use index if headers are requested
+            // Close for the time accessing the index
+            smalMailAccess.closeDelegateMailAccess();
 
+            // Access index
+            IndexAccess<MailMessage> indexAccess = null;
             try {
-                submitFolderJob(folder);
-            } catch (final OXException e) {
-                LOG.warn("Could not schedule folder job for folder {}.", folder, e);
+                final MailFields mfs = new MailFields(fields);
+                indexAccess = indexFacade.acquireIndexAccess(Types.EMAIL, session);
+                final boolean isIndexed = indexAccess.isIndexed(String.valueOf(accountId), folder);
+                if (isIndexed && MailUtility.getIndexableFields(indexAccess).containsAll(mfs)) {
+                    final AccountFolders accountFolders = new AccountFolders(String.valueOf(accountId), Collections.singleton(folder));
+                    final QueryParameters.Builder builder = new QueryParameters.Builder().setAccountFolders(Collections.singleton(accountFolders));
+
+                    if (sortField != null) {
+                        final MailField field = MailField.getField(sortField.getField());
+                        final MailIndexField indexSortField = MailIndexField.getFor(field);
+                        if (indexSortField != null) {
+                            builder.setSortField(indexSortField);
+                        }
+
+                        if (order != null) {
+                            builder.setOrder(order == OrderDirection.ASC ? Order.ASC : Order.DESC);
+                        }
+                    }
+
+                    final QueryParameters parameters = builder.setHandler(SearchHandlers.CUSTOM).setSearchTerm(searchTerm).build();
+                    final IndexResult<MailMessage> result = indexAccess.query(parameters, MailIndexField.getFor(fields));
+
+                    List<IndexDocument<MailMessage>> documents = result.getResults();
+                    List<MailMessage> mails;
+                    if (indexRange != null) {
+                        final int fromIndex = indexRange.start;
+                        int toIndex = indexRange.end;
+                        if ((documents == null) || documents.isEmpty()) {
+                            mails = Collections.emptyList();
+                        }
+                        if ((fromIndex) > documents.size()) {
+                            /*
+                             * Return empty iterator if start is out of range
+                             */
+                            mails = Collections.emptyList();
+                        }
+                        /*
+                         * Reset end index if out of range
+                         */
+                        if (toIndex >= documents.size()) {
+                            toIndex = documents.size();
+                        }
+                        documents = documents.subList(fromIndex, toIndex);
+                    }
+
+                    mails = IndexDocumentHelper.messagesFrom(documents);
+                    return mails.toArray(new MailMessage[mails.size()]);
+                }
+            } catch (final Throwable t) {
+                ExceptionUtils.handleThrowable(t);
+                LOG.warn("Index search failed. Falling back to message storage.", t);
+            } finally {
+                if (indexAccess != null) {
+                    indexFacade.releaseIndexAccess(indexAccess);
+                }
+
+                try {
+                    submitFolderJob(folder);
+                } catch (final OXException e) {
+                    LOG.warn("Could not schedule folder job for folder {}.", folder, e);
+                }
             }
         }
 
         // Fallback to message storage
-        return smalMailAccess.getDelegateMailAccess().getMessageStorage().searchMessages(folder, indexRange, sortField, order, searchTerm, fields);
+        IMailMessageStorage messageStorage = smalMailAccess.getDelegateMailAccess().getMessageStorage();
+        if (messageStorage instanceof IMailMessageStorageExt) {
+            IMailMessageStorageExt storageExt = (IMailMessageStorageExt) messageStorage;
+            return storageExt.searchMessages(folder, indexRange, sortField, order, searchTerm, fields, headerNames);
+        }
+
+        // Load headers manually
+        if (null == headerNames || headerNames.length <= 0) {
+            return messageStorage.searchMessages(folder, indexRange, sortField, order, searchTerm, fields);
+        }
+
+        MailField[] fieldsToUse = fields;
+        fieldsToUse = MailFields.addIfAbsent(fieldsToUse, MailField.ID);
+        MailMessage[] results = messageStorage.searchMessages(folder, indexRange, sortField, order, searchTerm, fieldsToUse);
+        if (null == results || results.length <= 0) {
+            return results;
+        }
+
+        int length = results.length;
+        MailMessage[] headers;
+        {
+            String[] ids = new String[length];
+            for (int i = ids.length; i-- > 0;) {
+                MailMessage m = results[i];
+                ids[i] = null == m ? null : m.getMailId();
+            }
+            headers = messageStorage.getMessages(folder, ids, MailFields.toArray(MailField.HEADERS));
+        }
+
+        for (int i = length; i-- > 0;) {
+            MailMessage mailMessage = results[i];
+            if (null != mailMessage) {
+                MailMessage header = headers[i];
+                if (null != header) {
+                    for (String headerName : headerNames) {
+                        String[] values = header.getHeader(headerName);
+                        if (null != values) {
+                            for (String value : values) {
+                                mailMessage.addHeader(headerName, value);
+                            }
+                        }
+                    }
+                }
+            }
+        }
+
+        return results;
     }
 
     @Override
@@ -336,11 +436,10 @@ public final class SmalMessageStorage extends AbstractSMALStorage implements IMa
     @Override
     public MailMessage getMessage(final String folder, final String mailId, final boolean markSeen) throws OXException {
         final MailMessage mail = smalMailAccess.getDelegateMailAccess().getMessageStorage().getMessage(folder, mailId, markSeen);
-        if (mail == null)  {
-            throw MailExceptionCode.MAIL_NOT_FOUN_BY_MESSAGE_ID.create(folder, mailId);
+        if (mail != null)  {
+            mail.setAccountId(accountId);
         }
 
-        mail.setAccountId(accountId);
         // TODO: this may be critical to performance.
 //        try {
 //            if (!mail.isPrevSeen()) {
@@ -573,4 +672,26 @@ public final class SmalMessageStorage extends AbstractSMALStorage implements IMa
         return builder;
     }
 
+    @Override
+    public MailMessage[] getMessagesByMessageIDByFolder(String fullName, String... messageIDs) throws OXException {
+        final IMailMessageStorage messageStorage = smalMailAccess.getDelegateMailAccess().getMessageStorage();
+        if (messageStorage instanceof IMailMessageStorageExt) {
+            final IMailMessageStorageExt messageStorageExt = (IMailMessageStorageExt) messageStorage;
+            return messageStorageExt.getMessagesByMessageID(messageIDs);
+        }
+        final SearchTerm<?> searchTerm;
+        if (1 == messageIDs.length) {
+            searchTerm = new com.openexchange.mail.search.HeaderTerm("Message-ID", messageIDs[0]);
+        } else {
+            return EMPTY_RETVAL;
+        }
+        return messageStorage.searchMessages(
+            fullName,
+            IndexRange.NULL,
+            MailSortField.RECEIVED_DATE,
+            OrderDirection.ASC,
+            searchTerm,
+            FIELDS_ID_AND_FOLDER);
+    }
+
 }
